import { useCallback, useEffect, useMemo, useRef } from "react";
import { useQuery, useQueryClient } from "react-query";
import { useFirestore } from "./Provider";
import { collectionCache } from "./Cache";
import { empty } from "./types";

const createFirestoreRef = (firestore, path, {
  where,
  orderBy,
  limit,
  startAt,
  endAt,
  startAfter,
  endBefore,
  isCollectionGroup
}) => // { isCollectionGroup = false }: { isCollectionGroup?: boolean } = empty.object
{
  let ref = firestore.collection(path);

  if (isCollectionGroup) {
    ref = firestore.collectionGroup(path);
  }

  if (where) {
    function multipleConditions(w) {
      return !!w && Array.isArray(w[0]);
    }

    if (multipleConditions(where)) {
      where.forEach(w => {
        ref = ref.where(w[0], w[1], w[2]);
      });
    } else if (typeof where[0] === "string" && typeof where[1] === "string") {
      ref = ref.where(where[0], where[1], where[2]);
    }
  }

  if (orderBy) {
    if (typeof orderBy === "string") {
      ref = ref.orderBy(orderBy);
    } else if (Array.isArray(orderBy)) {
      function multipleOrderBy(o) {
        return Array.isArray(o[0]);
      }

      if (multipleOrderBy(orderBy)) {
        orderBy.forEach(([order, direction]) => {
          ref = ref.orderBy(order, direction);
        });
      } else {
        const [order, direction] = orderBy;
        ref = ref.orderBy(order, direction);
      }
    }
  }

  if (startAt) {
    ref = ref.startAt(startAt);
  }

  if (endAt) {
    ref = ref.endAt(endAt);
  }

  if (startAfter) {
    ref = ref.startAfter(startAfter);
  }

  if (endBefore) {
    ref = ref.endBefore(endBefore);
  }

  if (limit) {
    ref = ref.limit(limit);
  }

  return ref;
};

const createListenerAsync = async (firestore, queryClient, path, queryString) => {
  return new Promise(resolve => {
    var _JSON$parse;

    const query = (_JSON$parse = JSON.parse(queryString)) !== null && _JSON$parse !== void 0 ? _JSON$parse : {};
    const ref = createFirestoreRef(firestore, path, query);
    const unsubscribe = ref.onSnapshot({
      includeMetadataChanges: true
    }, querySnapshot => {
      const data = [];
      querySnapshot.forEach(doc => {
        var _doc$data;

        const docData = (_doc$data = doc.data()) !== null && _doc$data !== void 0 ? _doc$data : empty.object;
        const docToAdd = { ...docData,
          id: doc.id,
          exists: doc.exists,
          hasPendingWrites: doc.metadata.hasPendingWrites
        }; // update individual docs in the cache

        queryClient.setQueryData(doc.ref.path, docToAdd);
        data.push(docToAdd);
      }); // resolve initial data

      resolve({
        initialData: data,
        unsubscribe
      }); // update on listener fire

      queryClient.setQueryData([path, queryString], data);
    });
  });
};
/**
 * Call a Firestore Collection
 * @template Doc
 * @param path String if the document is ready. If it's not ready yet, pass `null`, and the request won't start yet.
 * @param [query] - Dictionary with options to query the collection.
 * @param [options] - takes any of useQuery options.
 */


export const useCollection = (path, options, query) => {
  const {
    firestore
  } = useFirestore();
  const queryClient = useQueryClient();
  const unsubscribeRef = useRef(null);
  const {
    where,
    endAt,
    endBefore,
    startAfter,
    startAt,
    orderBy,
    limit,
    // __unstableCollectionGroup: isCollectionGroup = false,
    isCollectionGroup
  } = query || {}; // why not just put this into the ref directly?
  // so that we can use the useEffect down below that triggers revalidate()

  const memoQueryString = useMemo(() => JSON.stringify({
    where,
    endAt,
    endBefore,
    startAfter,
    startAt,
    orderBy,
    limit,
    isCollectionGroup
  }), [endAt, endBefore, isCollectionGroup, limit, orderBy, startAfter, startAt, where]);

  async function fetch() {
    const {
      unsubscribe,
      initialData
    } = await createListenerAsync(firestore, queryClient, path, memoQueryString);
    unsubscribeRef.current = unsubscribe;
    return initialData;
  }

  const {
    data,
    status,
    error
  } = useQuery([path, memoQueryString], fetch, { ...options,
    notifyOnChangeProps: "tracked"
  });
  useEffect(() => {
    //should it go before the useQuery?
    return () => {
      // clean up listener on unmount if it exists
      if (unsubscribeRef.current) {
        unsubscribeRef.current();
        unsubscribeRef.current = null;
      }
    }; // should depend on the path, queyr being the same...
  }, [path, memoQueryString]); // add the collection to the cache,
  // so that we can mutate it from document calls later

  useEffect(() => {
    if (path) collectionCache.addCollectionToCache(path, memoQueryString);
  }, [path, memoQueryString]);
  /**
   * `add(data)`: Extends the Firestore document [`add` function](https://firebase.google.com/docs/firestore/manage-data/add-data).
   * - It also updates the local cache using react-query's `setQueryData`. This will prove highly convenient over the regular `add` function provided by Firestore.
   */

  const add = useCallback(newData => {
    if (!path) return null;
    const dataArray = Array.isArray(newData) ? newData : [newData];
    const ref = firestore.collection(path);
    const docsToAdd = dataArray.map(doc => ({ ...doc,
      // generate IDs we can use that in the local cache that match the server
      id: ref.doc().id
    })); // add to network

    const batch = firestore.batch();
    docsToAdd.forEach(({
      id,
      ...doc
    }) => {
      // take the ID out of the document
      batch.set(ref.doc(id), doc);
    });
    return batch.commit();
  }, [path, firestore]);
  return {
    data,
    status,
    error,
    add,

    /**
     * A function that, when called, unsubscribes the Firestore listener.
     *
     * The function can be null, so make sure to check that it exists before calling it.
     *
     * Note: This is not necessary to use. `useCollection` already unmounts the listener for you. This is only intended if you want to unsubscribe on your own.
     */
    unsubscribe: unsubscribeRef.current
  };
};
//# sourceMappingURL=useCollection.js.map